local dropdashchar = function(player)
	return (player.charability == CA_DOUBLEJUMP and player.charability2 == CA2_SPINDASH)
end

addHook("PreThinkFrame", function()
	for player in players.iterate()
		local mo = player.mo
		if not (mo and dropdashchar(player)) then 
			player.dropswap = 0
		continue end
		
		//Initialize
		if player.dropswap == nil then
			player.dropswap = 0
		end
		
		local jump = player.cmd.buttons&BT_JUMP
		local spin = player.cmd.buttons&BT_USE
		local jumplast = player.pflags&PF_JUMPDOWN
		local spinlast = player.pflags&PF_USEDOWN
		local lastswap = player.dropswap		
		local shieldtype = player.powers[pw_shield]&~(SH_PITY|SH_PINK|SH_STACK) and not(player.powers[pw_super])
		
		if P_IsObjectOnGround(mo) then
			player.dropswap = 0
		end
		if player.pflags&PF_JUMPED
			//Holding jump key
			if (jumplast) and not(spin)
				player.dropswap = 0
			end
			//Use shield ability on jump if secondary jump is exhausted
			if (jump and not jumplast and player.secondjump and shieldtype)
				player.dropswap = 1
				player.pflags = $&~PF_USEDOWN
				player.cmd.buttons = $|BT_USE
			end
		end
	end
end)

addHook("ShieldSpecial", function(player)
	local mo = player.mo
	if dropdashchar(player) and not(player.dropswap) then return true end
	player.secondjump = 0
end)

addHook("SpinSpecial", function(player)
	local mo = player.mo
	//Drop dash
	local shieldtype = player.powers[pw_shield]&~(SH_PITY|SH_PINK|SH_STACK)
	if dropdashchar(player) and player.pflags&PF_JUMPED and not(player.pflags&PF_USEDOWN)
	and not(shieldtype and player.dropswap)
		mo.dropprep = 1
		mo.dropmomz = 0
	return true end
	if player.secondjump and shieldtype
		player.pflags = $&~PF_THOKKED
	end
end)

local StopDropDash = function(player)
	local mo = player.mo
	if not(mo.dropprep) then return end
	mo.dropprep = 0
	if player.pflags&(PF_SPINNING) or (player.pflags&PF_JUMPED and not(player.pflags&PF_NOJUMPDAMAGE))
		mo.state = S_PLAY_ROLL
	end
end


addHook("AbilitySpecial",function(player)
	player.mo.dropprep = min(1,$)
end)

addHook("ThinkFrame",function(player)
	for player in players.iterate()
		if player.mo.dropprep == nil then
			player.mo.dropprep = 0
			player.mo.dropmomz = 0
		end
		if P_PlayerInPain(player) or not(player.playerstate == PST_LIVE) or player.exiting or player.powers[pw_nocontrol]
			StopDropDash(player)
		end
		local mo = player.mo

		//Cancel dropdash
		if (not(player.cmd.buttons & BT_USE) and mo.dropprep)
		or mo.eflags & MFE_SPRUNG
		or P_PlayerInPain(player)
		or (not (mo.eflags & MFE_JUSTHITFLOOR) and not((player.pflags&PF_JUMPED)&~PF_NOJUMPDAMAGE))
			StopDropDash(player)
		return end
		
		//Handle dropdash
		if player.charability == CA_DOUBLEJUMP and player.charability2 == CA2_SPINDASH 
			if mo.dropprep
				and not(mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS1+5)
				//Air state
				if not (mo.eflags & MFE_JUSTHITFLOOR)
					//Charge dropdash
					if player.cmd.buttons & BT_USE
						if mo.dropprep < 100
							mo.dropprep = $+100/10
							if mo.dropprep >= 100
								S_StartSound(mo, sfx_drpdsh, player)
								mo.dropprep = 100
								mo.state = S_DROPDASH1
							end
						end
						//Fully charged
						if mo.dropprep == 100
							mo.dropmomz = mo.momz
							mo.thoklayer = P_SpawnGhostMobj(mo)
							mo.thoklayer.fuse = 5
							mo.thoklayer.state = S_THOK
						end
					end
				//Grounded state
				else
					//Execute dropdash
					if player.cmd.buttons & BT_USE and mo.dropprep == 100
						mo.state = S_PLAY_ROLL
						player.pflags = $1|PF_SPINNING
						S_StartSound(mo, sfx_zoom, player)
						local speed = abs(player.mo.dropmomz/2)+FixedHypot(player.rmomx,player.rmomy)
						local actionspd = player.actionspd
						if player.powers[pw_super] then
							actionspd = $*3/2
						end
						if mo.eflags&MFE_UNDERWATER then actionspd = $*2/3 end
						P_InstaThrust(mo, mo.angle,max(actionspd,speed))
						
						//Dust
						local function r(mo,value)
							local w = mo.radius/FRACUNIT/2
							return value+P_RandomRange(-w,w)*FRACUNIT
						end
						for n = 1,4
							local d = P_SpawnMobj(r(mo,mo.x),r(mo,mo.y),mo.z,MT_SPINDUST)
							d.extravalue1 = FRACUNIT
							d.extravalue2 = FRACUNIT
							if P_MobjFlip(mo) == -1 then
								d.eflags = $|MFE_VERTICALFLIP
							end
							P_SetObjectMomZ(d,FRACUNIT*2)
							local l = 90*n/4
							P_InstaThrust(d,mo.angle+ANGLE_135+l*ANG1,actionspd/2)
						end
					end
					//Reset
					StopDropDash(player)
				end
			end
		end
	end
end)